from timeit import timeit

import geopy.distance
import mercantile
import pytest

import georgio

from .locations import x_y_z_extensions

@pytest.mark.parametrize("x, y, z, extension_meters", x_y_z_extensions)
def test_wm_tile_expanded_bbox(x, y, z, extension_meters):
    west, south, east, north = georgio.wm_tile_expanded_bbox(x, y, z, extension_meters)
    tile_west, tile_south, tile_east, tile_north = mercantile.bounds(x, y, z)
    assert west < tile_west
    assert south < tile_south
    assert east > tile_east
    assert north > tile_north

    center_lon = (west + east) / 2
    center_lat = (north + south) / 2
    tile_center_lon = (tile_west + tile_east) / 2
    tile_center_lat = (tile_north + tile_south) / 2

    assert center_lon == pytest.approx(tile_center_lon)
    assert center_lat == pytest.approx(tile_center_lat)

    tile_corner_to_corner = geopy.distance.great_circle((tile_north, tile_west), (tile_south, tile_east)).m
    tile_extended_corner_to_corner = tile_corner_to_corner + (2 * extension_meters)
    bbox_corner_to_corner = geopy.distance.great_circle((north, west), (south, east)).m

    assert bbox_corner_to_corner >= tile_extended_corner_to_corner

    tile_corner_to_bbox_corner = geopy.distance.great_circle((tile_north, tile_east), (north, east)).m
    assert tile_corner_to_bbox_corner >= extension_meters

@pytest.mark.parametrize("x, y, z, extension_meters", x_y_z_extensions)
def test_wm_tile_expanded_bbox_benchmark(x, y, z, extension_meters):
    run_time = timeit(
        f"wm_tile_expanded_bbox({x}, {y}, {z}, {extension_meters})",
        setup="from georgio import wm_tile_expanded_bbox",
        number=100_000,
    )
    print("Average run time: {} ns".format(run_time/100_000*1_000_000_000))



from datetime import datetime

if __name__ == "__main__":
    now = datetime.utcnow()
    julian_int = int(now.strftime('%j'))
    minute_of_day = now.hour  * 60 + now.minute
    print('{}.{}.{}'.format(now.year, julian_int, minute_of_day))



from itertools import permutations
from timeit import timeit

import pytest

import geopy.distance
import georgio

from .locations import lon_lats

def test_great_circle_distance():
    for pair in permutations(lon_lats, 2):
        lon1 = pair[0][0]
        lat1 = pair[0][1]
        lon2 = pair[1][0]
        lat2 = pair[1][1]
        dist1 = georgio.great_circle_distance_with_radius(lon1, lat1, lon2, lat2, geopy.distance.EARTH_RADIUS*1000)
        dist2 = geopy.distance.great_circle((lat1, lon1), (lat2, lon2)).m
        assert dist1 == pytest.approx(dist2)

def test_great_circle_benchmark():
    georgio_results = []

    for pair in permutations(lon_lats, 2):
        lon1 = pair[0][0]
        lat1 = pair[0][1]
        lon2 = pair[1][0]
        lat2 = pair[1][1]
        run_time = timeit(
            f'great_circle_distance({lon1}, {lat1}, {lon2}, {lat2})',
            setup='from georgio import great_circle_distance',
            number=10_000,
        )
        georgio_results.append(run_time/10_000*1_000_000_000)

    print(f'georgio fastest run: {min(georgio_results)} ns, slowest run: {max(georgio_results)} ns')

    geopy_results = []

    for pair in permutations(lon_lats, 2):
        lon1 = pair[0][0]
        lat1 = pair[0][1]
        lon2 = pair[1][0]
        lat2 = pair[1][1]
        run_time = timeit(
            f'great_circle(({lat1}, {lon1}), ({lat2}, {lon2})).m',
            setup='from geopy.distance import great_circle',
            number=10_000,
        )
        geopy_results.append(run_time/10_000*1_000_000_000)

    print(f'geopy fastest run: {min(geopy_results)} ns, slowest run: {max(geopy_results)} ns')

    # makes sure georgio's slowest run is faster than geopy's fastest run
    assert max(georgio_results) < min(geopy_results)



from timeit import timeit

import mercantile
import numba
import numpy as np
import pytest

import georgio

from .locations import lon_lats

@numba.njit(fastmath=True)
def jit_ul(xtile, ytile, zoom):
    inv_z2 = 1 / 2 ** zoom
    return (
        # lon_deg
        xtile * inv_z2 * 360.0 - 180.0,
        # lat_deg
        np.degrees(np.arctan(np.sinh(np.pi * (1 - 2 * ytile * inv_z2)))),
    )

@numba.njit(fastmath=True)
def jit_bounds(xtile, ytile, zoom, epsilon=1.0e-9):
    a = jit_ul(xtile, ytile, zoom)
    b = jit_ul(xtile + 1, ytile + 1, zoom)
    return a[0], b[1] + epsilon, b[0] - epsilon, a[1]  # west, south, east, north

def test_wm_upper_left():
    jit_lon, jit_lat = jit_ul(1205, 1540, 12)
    georgio_lon, georgio_lat = georgio.wm_upper_left(1205, 1540, 12)
    assert georgio_lon == pytest.approx(jit_lon)
    assert georgio_lat == pytest.approx(jit_lat)

def test_wm_upper_left_benchmark():
    georgio_results = []

    for lon_lat in lon_lats:
        for zoom in range(2, 20):
            tile = mercantile.tile(lon_lat[0], lon_lat[1], zoom)
            run_time = timeit(
                f'wm_upper_left({tile.x}, {tile.y}, {tile.z})',
                setup='from georgio import wm_upper_left',
                number=10_000,
            )
            georgio_results.append(run_time/10_000*1_000_000_000)

    print(f'georgio fastest run: {min(georgio_results)} ns, slowest run: {max(georgio_results)} ns')

    jit_results = []

    for lon_lat in lon_lats:
        for zoom in range(2, 20):
            tile = mercantile.tile(lon_lat[0], lon_lat[1], zoom)
            run_time = timeit(
                f'jit_ul({tile.x}, {tile.y}, {tile.z})',
                globals=globals(),
                number=10_000,
            )
            jit_results.append(run_time/10_000*1_000_000_000)

    print(f'jit fastest run: {min(jit_results)} ns, slowest run: {max(jit_results)} ns')

    # makes sure georgio's slowest run is faster than jit's fastest run
    #assert max(georgio_results) < min(jit_results)

def test_wm_bounds():
    j_west, j_south, j_east, j_north = jit_bounds(1205, 1540, 12)
    g_west, g_south, g_east, g_north = georgio.wm_bounds(1205, 1540, 12)
    assert g_west == pytest.approx(j_west)
    assert g_south == pytest.approx(j_south)
    assert g_east == pytest.approx(j_east)
    assert g_north == pytest.approx(j_north)

def test_wm_bounds_benchmark():
    georgio_results = []

    for lon_lat in lon_lats:
        for zoom in range(2, 20):
            tile = mercantile.tile(lon_lat[0], lon_lat[1], zoom)
            run_time = timeit(
                f'wm_bounds({tile.x}, {tile.y}, {tile.z})',
                setup='from georgio import wm_bounds',
                number=10_000,
            )
            georgio_results.append(run_time/10_000*1_000_000_000)

    print(f'georgio fastest run: {min(georgio_results)} ns, slowest run: {max(georgio_results)} ns')

    jit_results = []

    for lon_lat in lon_lats:
        for zoom in range(2, 20):
            tile = mercantile.tile(lon_lat[0], lon_lat[1], zoom)
            run_time = timeit(
                f'jit_bounds({tile.x}, {tile.y}, {tile.z})',
                globals=globals(),
                number=10_000,
            )
            jit_results.append(run_time/10_000*1_000_000_000)

    print(f'jit fastest run: {min(jit_results)} ns, slowest run: {max(jit_results)} ns')

    # makes sure georgio's slowest run is faster than jit's fastest run
    #assert max(georgio_results) < min(jit_results)
